/* compjuga.c
 *
 * Created: Nov. 13, 1998 by Daniel M. German (dmg@csg.uwaterloo.ca)
 * Revised: 
 * Copyright 1998 Daniel M. German (dmg@csg.uwaterloo.ca)
 *
 * This file is part of compjuga, a Spanish verbs conjugator 
 * (see http://csg.uwaterloo.ca/~dmg/compjuga
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: compjuga.c,v 1.1 1998/11/18 20:43:33 dmg Exp $
 * 
 */

#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>
#include <gdbm.h>

#include "compjuga.h"

#ifdef NDEBUG
#define DEBUG(a)
#define DEBUG_2(a,b)
#define DEBUG_3(a,b,c)
#define DEBUG_4(a,b,c,d)
#else
#define DEBUG(a)  fprintf(stderr, a);
#define DEBUG_2(a,b)  fprintf(stderr, a,b);
#define DEBUG_3(a,b,c) fprintf(stderr, a,b,c);
#define DEBUG_4(a,b,c,d)  fprintf(stderr, a,b,c,d);
#endif
/* The conjugation engine will receive a verb as a parameter, and the 
   tense and person required */

#define CHARACTER_SET   "abcdefghijklmnopqrstuvwxyz"

#define YO_PERSON           0x01
#define TU_PERSON           0x02
#define EL_PERSON           0x04
#define NOSOTROS_PERSON     0x08
#define USTEDES_PERSON      0x10
#define ELLOS_PERSON        0x20


#define PRESENT_TENSE_OFFSET        0
#define PRETERITO_TENSE_OFFSET      1
#define FUTURO_TENSE_OFFSET         2
#define COPRETERITO_TENSE_OFFSET    3
#define POSPRETERITO_TENSE_OFFSET   4
#define SUBJUNTIVO_PRESENTE_TENSE_OFFSET 5
#define SUBJUNTIVO_PRETERITO_TENSE_OFFSET 6
#define SUBJUNTIVO_PRETERITO_2_TENSE_OFFSET 7
#define SUBJUNTIVO_FUTURO_2_TENSE_OFFSET 8
#define IMPERATIVO_TENSE_OFFSET          9
#define IMPERSONALES_TENSE_OFFSET       10


#define GERUNDIO_OFFSET       (IMPERSONALES_TENSE_OFFSET * 6 + 0)
#define PARTICIPIO_OFFSET     (IMPERSONALES_TENSE_OFFSET * 6 + 1)
#define PARTICIPIO_COMP_OFFSET     (IMPERSONALES_TENSE_OFFSET * 6 + 2)


/* Errors */

#define ERROR_NOT_A_VERB           -1
#define ERROR_INVALID_CHARACTER    -2
#define ERROR_NOT_IN_DATABASE      -3


char* PersonaNames[] = {"1", "2", "3", "1", "2", "3"}
;

char* TenseNames[] = 
{
  //  "Presente","Pasado, Pretrito Indefinido o Pretrito", "Futuro", 
  //  "Pretrito Imperfecto o Copretrito",
  //  "Condicional o Pospretrito",
   "Presente","Pasado", "Futuro", 
   "Copretrito",
   "Pospretrito",
   "Presente Subjuntivo",
   "Pasado Subjuntivo",
   "Pasado Subjuntivo (segunda variacin)",
   "Futuro Subjuntivo",
};

int iError=0;

GDBM_FILE dbf;



int Verify_It_Looks_Like_A_Verb(char *szVerb)
{
  int iLen = strlen(szVerb);
  int i;
  /*
    Return 0 if szVerb looks like  a verb of error otherwise 
   */

  /* A verb looks like a verb is uses a character set 
     from Spanish, and it ends in ar,er or ir
  */
     

  /* change to lowercase */
  for (i=0;*(szVerb+i) != (char)NULL;i++)
    {
      *(szVerb+i) = tolower(*(szVerb+i));
      switch (*(szVerb+i))
	{
	case '': *(szVerb+i) = '';break;
	case '': *(szVerb+i) = '';break;
	case '': *(szVerb+i) = '';break;
	case '': *(szVerb+i) = '';break;
	case '': *(szVerb+i) = '';break;
	case '': *(szVerb+i) = '';break;
	case '': *(szVerb+i) = '';break;
	}
    }

  /* It assumes the verb is in lowercase */
  if (strspn(szVerb, CHARACTER_SET) != iLen)
    {
      return ERROR_INVALID_CHARACTER;
    }
  /* By now it should be lowercase and only Spanish characters */
  
  /* Verify its ending */
  
  /* For the time being, only handle "ar, er, ir"  */
  if (szVerb[iLen-1] != 'r' ||
      strpbrk(&szVerb[iLen-2], "aei") == NULL)
    {
      return ERROR_NOT_A_VERB;
    }
  return 0;
}


int Get_Family_From_Verb_Record(datum *pRecord)
{
  return ((unsigned char)(*(pRecord->dptr)));
}


int Get_Number_Irreg_Conj_From_Verb_Record(datum *pRecord)
{
  DEBUG_2("Number of irreg records %d", (unsigned char)(*(pRecord->dptr+1)));
  return (unsigned char)(*(pRecord->dptr+1));
}

datum Get_Verb_Record(char *szVerb)
{
  /* First verify that it is actually a verb */
  datum Key;
  datum Verb;

  Key.dptr = szVerb;
  Key.dsize= strlen(szVerb);
  
  Verb = gdbm_fetch(dbf,Key);
  
  return Verb;
}



void Display_Error(char *szMessage)
{
  fprintf(stderr, szMessage);
}

void Init_Conjugator(void)
{
  /* Open data file */
  printf("Comjuga Version Alpha %3.2f\n", VERSION);
  if ((dbf  =  gdbm_open  (DATA_FILE, 0, GDBM_READER, 0, 0)) == NULL)
    {
      Display_Error("Unable to open database file");
      exit(1);
    }
}

void End_Conjugator(void)
{
  /* Open data file */
   gdbm_close(dbf);
}




conjugation_irreg_spec *Load_Conjugation_Spec_From_Datum(datum Record)
{
  /* The format for the table is: */
  /*

   Family                  unsigned char
   Number of conjugationns unsigned char
   Foreach conjugation:    sequence of null terminated strings 
      ItemIndex              unsigned char
      Offset                 unsigned char
      Replacement            
   Options
   Comments
  */
  int iNoItems = 0, i;
  char *cPtr;
  conjugation_irreg_spec *pSpec;

  assert(Record.dptr != NULL);
  
  iNoItems = (unsigned char)(*(Record.dptr+1));
  /* Allocate iNoItems records plus one (last is a terminator) */
  if ((pSpec = malloc(sizeof(conjugation_irreg_spec) * (iNoItems +1))) == NULL)
    return NULL;
      
  
  cPtr = Record.dptr+2;
  for (i=0;i<iNoItems;i++)
    {
        (pSpec+i)->iItemIndex = (unsigned char)*cPtr++;
	(pSpec+i)->iOffset = (unsigned char)*cPtr++;
	(pSpec+i)->szSuffix = strdup(cPtr);
	/* Not enough memory ? */
	if ((pSpec+i)->szSuffix == NULL)
	  return NULL;
	cPtr+=strlen((pSpec+i)->szSuffix)+1;
	DEBUG_4(" Item: [%d] [%d] %s", 
		(pSpec+i)->iItemIndex, (pSpec+i)->iOffset, (pSpec+i)->szSuffix);
      }
  DEBUG("\n");
  (pSpec+i) -> iItemIndex = -1;
  (pSpec+i) -> iOffset = 0;
  (pSpec+i) -> szSuffix = NULL;
  return pSpec;
}

void Free_Irreg_Conjugation_Table(conjugation_irreg_spec *pSpec)
{
  int i=0;
  while ((pSpec+i)->iItemIndex >= 0)
    {
      free((pSpec+i)->szSuffix);
      i++;
    }
  free(pSpec);
}


conjugation *Create_Conjugation_Item(int iCount, string Items[])
{
  conjugation* pConjugation = NULL;
  int i;
  
  //  DEBUG_2("Creating conjugating %d\n",sizeof(conjugation))
  pConjugation = (conjugation*)malloc(sizeof(conjugation));
  assert(pConjugation != NULL);

  pConjugation->iVariants=iCount;
  for (i=0;i<iCount;i++)
    {
      //       DEBUG_2("Creating conjugating item %s\n",Items[i]);
      strcpy(pConjugation->pszConj[i], Items[i]);
    }
  return pConjugation;
}

void Conjugate_Item(string Result, char *szVerb, char *szSuffix,
		    int iIndex)
{
  strcpy(Result, szVerb);
  strcpy(Result+strlen(szVerb)-iIndex,szSuffix);
}



conjugation_table *Conjugate_Regular_Verb(char *szVerb,int iFamily)
{
  /* Conjugate a regular verb */
  int i,j,iIndex,iFirst,iTemp;

  conjugation_table* pConj;
  conjugation_spec* pVerbSpec;
  string Items[3];

  assert(iFamily >= 3 && iFamily <= 5);
  /* We will save the conjugations here */

  /* Allocate memory for the data structure */

  //  DEBUG_2("Conjugating regular %d\n", iFamily);

  if ((pConj = (conjugation_table*) malloc(sizeof(conjugation_table))) == 
      (conjugation_table*)NULL)
      return (conjugation_table*)NULL;
  
  pVerbSpec = Regular_Specs[iFamily-3];

  for (i = 0;i<NO_CONJUGATIONS;i++)
    {
      /* if offset >> 4 is different from 0 then its value is the index in 
	 pVerbSpec to the other compjugation for this tense/person
      */
      j=0;
      iIndex = i;
      //  DEBUG_2("Conjugating regular %d\n",i);
      iFirst = 1;
      while ( iIndex !=0 || iFirst)
	{
	  iFirst = 0;
	  Conjugate_Item(Items[j],szVerb, pVerbSpec[iIndex].szSuffix, 
			 pVerbSpec[iIndex].iOffset);
	  //	  DEBUG_2("Conjugating regular %s\n",Items[j]);

	  iTemp = pVerbSpec[iIndex].iOffset >> 4;
	  if (iTemp > 0)
	    iIndex +=iTemp;
	  else
	    iIndex = 0;
	  j++;
	}
      (*pConj)[i] = Create_Conjugation_Item(j,Items);
      //      DEBUG_2("Conjugating regular %d\n",j);

    }
  //  DEBUG("Ending conjugating regular\n");
  return pConj;
}

int Verb_Group(char *szVerb)
{
  /* returns 1 for -ar verbs, 2 for -er verbs, 3 for -ir verbs */
  switch (szVerb[strlen(szVerb)-2])
    {
    case 'a':
      return 3;
    case 'e':
      return 4;
    case 'i':
    case '':
      return 5;
    }
  assert(1);
  return 0;
}


conjugation_table *Conjugate_Irregular_Verb(char *szVerb,datum *pVerb)
{
  /* Conjugate a regular verb */
  int iItemIndex;
  conjugation *pConjugation;
  conjugation_table* pConj;
  conjugation_irreg_spec *pIrregConj,*pSaveIrregConj;
  string szTemp;

  //  DEBUG("We are to conjugate it as regular \n");
  pConj = Conjugate_Regular_Verb(szVerb, Verb_Group(szVerb));
  //  DEBUG("We have conjugated it as regular \n");

  pIrregConj = pSaveIrregConj = Load_Conjugation_Spec_From_Datum(*pVerb);

  if (Get_Number_Irreg_Conj_From_Verb_Record(pVerb) == 0)
    {
      switch(Get_Family_From_Verb_Record(pVerb))
	{
	case 71:
	  pIrregConj = CarTable;
	  break;
	case 72:
	  pIrregConj = GarTable;
	  break;
	case 73:
	  pIrregConj = ZarTable;
	  break;
	default:
	  fprintf(stderr, " Illegal record for verb %s\n", szVerb);
	  exit(1);
	}
    }
       


  while (pIrregConj -> iItemIndex >= 0)
    {
      iItemIndex = pIrregConj -> iItemIndex;

      DEBUG_3("Conjugation [%d][%d] ", iItemIndex, iItemIndex % 100);
      /* The 100 is needed because we mark optional conjugations
	 as follows above 100,hence index = iItemIndex % 100 */
      pConjugation = (*pConj)[iItemIndex%100]; 
      strcpy(szTemp, pConjugation->pszConj[0]);
      Conjugate_Item(szTemp,szVerb, pIrregConj->szSuffix, 
			 pIrregConj->iOffset);

      DEBUG_2("[%s] ", szTemp);

      if (iItemIndex >= 100)
	{
	  pConjugation->iVariants ++;
	  DEBUG_2("Added at [%d]", iItemIndex / 100);
	  strcpy(pConjugation->pszConj[iItemIndex/100],szTemp);
	}
      else
	{
	  strcpy(pConjugation->pszConj[0],szTemp);
	  /* This is a hack and it is ugly,
	     we have two entries for participio, for those verbs
	     that have 2 different ones, one for the 
	     compuestos, soooo, since the items are sorted,
	     we can assume that an irregular particio
	     affects the compuesto also. If there is a second
	     participio, it will be modified by its own record */
	  if (iItemIndex == PARTICIPIO_OFFSET)
	    strcpy((*pConj)[PARTICIPIO_COMP_OFFSET]->pszConj[0],szTemp);


	}
      pIrregConj++;
      DEBUG("\n");
    }
  //  Free_Irreg_Conjugation_Table(pIrregConj);
  
  return pConj;
}




conjugation_table *Conjugate_Verb(char *szVerb)
{
  int iFamilyVerb; 
  datum VerbRecord;

  if ((iError = Verify_It_Looks_Like_A_Verb(szVerb)) != 0)
    return NULL;

  VerbRecord = Get_Verb_Record(szVerb);

  // Verify that we got a result
  if (VerbRecord.dptr == NULL)
    {
      iError = ERROR_NOT_IN_DATABASE;
      return NULL;
    }
  iFamilyVerb = Get_Family_From_Verb_Record(&VerbRecord);
  DEBUG_2("Family: %d\n", iFamilyVerb);
  assert (iFamilyVerb > 0 && iFamilyVerb <100);
  
  /* There are some verbs that are family 3, 4, and 5,
     but have irregular impersonal tenses:
     abrir, so we have to check if it has an irreg
     record */

  if (iFamilyVerb >= 3 && iFamilyVerb <=5 &&
      Get_Number_Irreg_Conj_From_Verb_Record(&VerbRecord)==0)
    {
      /* Regular verb */
      return Conjugate_Regular_Verb(szVerb,iFamilyVerb);
    }
  else
    {
      return Conjugate_Irregular_Verb(szVerb,&VerbRecord);
    }
  return NULL;
}

void Free_Conjugation(conjugation_table *pConj)
{
  int i;
  if (pConj == NULL)
    return;

  for (i = 0;i<NO_CONJUGATIONS;i++)
    {
      free((*pConj)[i]);
    }
  free(pConj);
}


char *Substitute(char *szWhere, char* szWhat, char *szWith)
{
  /* Replace in string szWhere, the substring What, with With 
     once */

  char *ptr;
  int iLenWith = strlen(szWith);
  int iLenWhat  = strlen(szWhat);
  int iLenWhere = strlen(szWhere);
  
  while ((ptr = strstr(szWhere,szWhat)) != (char*)NULL)
    {
      memmove(ptr+iLenWith, ptr+iLenWhat, iLenWhere - (int)(ptr - szWhere) - iLenWhat+1);
      memmove(ptr, szWith, iLenWith);

      assert(strlen(szWhere) == iLenWhere - iLenWhat + iLenWith);
    }
  return szWhere;
}


void Print_Conjugations(char *szVerb, conjugation_table *pConjugation)
{
  int i,j,k;

  printf("Verbo: %s\n", szVerb);
  printf("Formas impersonales:\n  ");
  printf("Gerundio: %s ", (*pConjugation)[GERUNDIO_OFFSET]->pszConj[0]);
  printf("Particio: %s", (*pConjugation)[PARTICIPIO_OFFSET]->pszConj[0]);
  if (strcmp(
	     (*pConjugation)[PARTICIPIO_OFFSET]->pszConj[0],
	     (*pConjugation)[PARTICIPIO_COMP_OFFSET]->pszConj[0]) != 0)
    {
      
      printf(",%s ", (*pConjugation)[PARTICIPIO_COMP_OFFSET]->pszConj[0]);
      printf("Particio para tiempos compuestos: %s", 
	     (*pConjugation)[PARTICIPIO_COMP_OFFSET]->pszConj[0]);
    }
  printf("\n");
  
  for (i=PRESENT_TENSE_OFFSET;i<=SUBJUNTIVO_FUTURO_2_TENSE_OFFSET;i++)
    {
      printf("%s: \n  ", TenseNames[i]);
      for (k=0;k<6;k++)
	{
	  printf("%s: ", PersonaNames[k]);
	  for (j=0;j<(*pConjugation)[i*6+k]->iVariants;j++)
	    printf("%s ",(*pConjugation)[i*6+k]->pszConj[j]);
	}
      printf("\n");
    }
  printf("Imperativo:\n  ");
  
  i = IMPERATIVO_TENSE_OFFSET;
  
  for (k=1;k<6;k++) /* first person does not exist */
    {
      printf("%s: ", PersonaNames[k]);
      for (j=0;j<(*pConjugation)[i*6+k]->iVariants;j++)
	printf("%s ",(*pConjugation)[i*6+k]->pszConj[j]);
    }
  printf("\n");

}



void main (void)
{
  char s[100];
  conjugation_table *pConjugation;

  Init_Conjugator();

  while (1)
    {
      if (fgets(s,100,stdin) == NULL)
	break;
      
      s[strlen(s)-1] = 0;

      /* ok now we have to take care of potential accented characters */
      Substitute(s, "'i", "");
      Substitute(s, "'o", "");
      Substitute(s, "\"u", "");
      Substitute(s, "~n", "");

      pConjugation = Conjugate_Verb(s);
      if (pConjugation != NULL)
	{
	  Print_Conjugations(s,pConjugation);
	}
      else
	{
	  switch(iError)
	    {
	    case ERROR_NOT_IN_DATABASE:
	      printf("I don't know how to conjugate this verb [%s].\n", s);
	      break;
	    case ERROR_NOT_A_VERB:
	      printf("%s is not a verb.\n", s);
	      break;
	    case ERROR_INVALID_CHARACTER:
	      printf("%s has an invalid character.\n", s);
	      break;
	    }
	}
      
      Free_Conjugation(pConjugation);
      pConjugation = NULL;
    }

  End_Conjugator();

}
